"""
Renderers are used to serialize a response into specific media types.
They give us a generic way of being able to handle various media types
on the response, such as JSON encoded data or HTML output.
"""
import json
from xml.etree import ElementTree


class BaseRenderer(object):
    """
    All renderers should extend this class, setting the `media_type`
    and `format` attributes, and override the `.render()` method.
    """
    media_type = None
    format = None

    def render(self, data):
        raise NotImplementedError(
            'Renderer class requires .render() to be implemented'
        )


class JSONRenderer(BaseRenderer):
    """
    Renderer which serializes to JSON
    """
    media_type = 'application/json'
    format = 'json'

    def render(self, data):
        if data is None:
            return bytes()
        return json.dumps(data)


class XMLRenderer(BaseRenderer):
    """
    Renderer which serializes to XML
    """
    media_type = 'application/xml'
    format = 'xml'

    def _data_to_xml(self, data, tag=None):
        """
        Convert the input data to XML
        """
        tag = tag or 'response'
        elem = ElementTree.Element(tag)
        if isinstance(data, (list, tuple)):
            for item in data:
                elem.append(self._dict_to_xml(item, tag=tag or 'list'))
        elif isinstance(data, dict):
            for key, value in data.iteritems():
                elem.append(self._dict_to_xml(value, tag=key))
        else:
            elem.text = unicode(data)
        return elem

    def render(self, data, tag=None):
        if data is None:
            return bytes()
        return ElementTree.tostring(self._data_to_xml(data, tag))



"""
Contains shortcut functions to map a resource to URLs
"""
import webapp2
from webapp2_extras import routes
import http


def route(base_url, viewset, detail_field='id',
          allowed_list_methods=None, allowed_detail_methods=None):
    """
    Return a routes.PathPrefixRoute (See:)
    https://webapp-improved.appspot.com/guide/routing.html#path-prefix-routes
    mapping a request handler (`viewset`) with a url (`base_url`)

    The `viewset` must:
        - have `get()`, `post()`, `put()`, `delete()` defined
        or
        - extend `viewset.ModelViewSet`

    The `routes.PathPrefixRoute` has two routes:
        - The list view at `base_url`/ with URI: `base_url`-list
        - The detail view at `base_url`/<id> with URL `base_url`-detail

    Optional Parameters:
        - `detail_field`: The name of the field used to identify a resource
        - `allowed_list_methods`: List of HTTP methods allowed for list view.
                Default is [GET, POST]
        - `allowed_detail_methods`: List of HTTP methods allowed for
                detail view. Default is [GET, PUT, DELETE]
    """
    return routes.PathPrefixRoute(base_url, [
        webapp2.Route(
            '/',
            handler=viewset,
            name='{resource}-list'.format(resource=base_url),
            methods=allowed_list_methods or [http.GET, http.POST],
        ),
        webapp2.Route(
            '/<{detail_field}:\d+>'.format(detail_field=detail_field),
            handler=viewset,
            name='{resource}-detail'.format(resource=base_url),
            methods=allowed_detail_methods or [http.GET, http.PUT, http.DELETE]
        )
    ])



"""
Common HTTP constants.
Contains method names and status codes
"""

GET = 'GET'
POST = 'POST'
PUT = 'PUT'
PATCH = 'PATCH'
DELETE = 'DELETE'

HTTP_200_OK = 200
HTTP_201_CREATED = 201
HTTP_204_NO_CONTENT = 204
HTTP_400_BAD_REQUEST = 400
HTTP_401_UNAUTHORIZED = 401
HTTP_403_FORBIDDEN = 403
HTTP_404_NOT_FOUND = 404
HTTP_405_METHOD_NOT_ALLOWED = 405
HTTP_406_NOT_ACCEPTABLE = 406
HTTP_500_INTERNAL_SERVER_ERROR = 500



"""
Custom exceptions for Zenn-La
"""
from zennla import http


class APIException(Exception):
    """
    Base class for Zenn-La exceptions.
    Subclasses should provide `.status_code` and `.default_detail` properties.
    """
    status_code = http.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = {'detail': "A server error occurred."}

    def __init__(self, detail=None):
        msg = detail or self.default_detailn
        self.detail = msg if isinstance(
            msg, (dict, tuple, list)
        ) else {'details': msg}

    def __str__(self):
        return self.detail


class ImproperlyConfigured(APIException):
    """
    Raised when there is a problem with the configuration
    """
    status_code = http.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = {'detail': "Improperly Configured"}


class NonSerializableException(APIException):
    """
    Raised when a non-serializable object is received by a serializer
    """
    status_code = http.HTTP_400_BAD_REQUEST
    default_detail = {'detail': "Object is not serializable"}


class ValidationError(APIException):
    """
    Raised when the request made is not valid
    """
    status_code = http.HTTP_400_BAD_REQUEST
    default_detail = {'detail': "Invalid request"}


class UnacceptableRequest(APIException):
    """
    Raised when the request made is not acceptable
    """
    status_code = http.HTTP_406_NOT_ACCEPTABLE
    default_detail = {'detail': "The request made could not be accepted"}



"""
Serializers have a two-fold purpose:
    - To provide a way of translating ndb models into primitive data-types
        and transforming them into a JSON string
    - To validate and translate primitive data into ndb model fields
        and create/update an object
"""
from google.appengine.ext import ndb
from google.appengine.ext.db import BadValueError
from zennla.exceptions import NonSerializableException, ValidationError


class ModelSerializer(object):
    """
    ModelSerializer can be subclassed to create custom serializers
    for a model. The attributes that can be set are:
        - `model`: Set to a model class for which the serializers
                is being constructed.
        - `include_fields`: List of field names to be included in the
                serialized form of a model object.
        - `exclude_fields`: List of field names to be excluded
                during serialization. Takes precedence over `include_fields`
        - `translate_fields`: A dict mapping field names to the names used in
                serialized representation
    """
    include_fields = None
    exclude_fields = None
    translate_fields = {}
    model = None

    def _save(self, data, instance):
        """
        Populate and write an `instance` with `data` after validation
        Return the updated `instance`
        """
        validated_data = self._validate(data)
        instance.populate(**validated_data)
        if hasattr(self, 'pre_save'):
            self.pre_save(instance, data, validated_data)
        instance.put()
        if hasattr(self, 'post_save'):
            self.post_save(instance, data, validated_data)
        return instance

    def _validate(self, data, model=None):
        """
        Take `data` as the dict containing the input data
        Return a dictionary with the validated data
        Raise a validation error if invalid data is found
        """
        validated_data = {}
        model = model or self.model
        properties = model._properties

        # Perform field level validations
        for prop_name, prop in properties.iteritems():
            field_value = data.get(
                self.translate_fields.get(prop_name), data.get(prop_name)
            )
            if field_value is None:
                if prop._required:
                    raise ValidationError(
                        "Property {name} is required".format(name=prop_name)
                    )
                else:
                    validated_data[prop_name] = prop._default \
                        if not prop._repeated else []
            else:
                try:
                    if prop._repeated:
                        if not isinstance(field_value, (list, tuple)):
                            raise ValidationError(
                                "`{field}` should be a list.".format(
                                    field=prop_name
                                )
                            )
                        else:
                            validated_data[prop_name] = [
                                prop._validate(elem) or elem
                                for elem in field_value
                            ]
                    else:
                        validated_data[prop_name] = prop._validate(
                            field_value
                        ) or field_value
                except BadValueError as e:
                    raise ValidationError(str(e))

            # Perform any custom field validations
            field_validation_method = getattr(
                self, 'validate_' + prop_name, None
            )
            if field_validation_method is not None:
                field_validation_method(validated_data[prop_name])

        # Perform object level validation
        if hasattr(self, 'validate'):
            self.validate(validated_data, model)
        return validated_data

    def create(self, data, model=None):
        """
        Create a model instance with the given data
        Optionally take a `model` argument (default is self.model)
        """
        return self._save(data=data, instance=self.get_obj(model=model))

    def get_obj(self, id=None, model=None):
        """
        Return an instance at `id` of the given `model`
        (or of self.model if model isn't specified)
        Return a new instance if `id` isn't specified
        """
        model = model or self.model
        if not isinstance(model, ndb.model.MetaModel):
            raise ValidationError(
                "Expected an NDB model. Got {type} instead".format(
                    type=type(model).__name__
                )
            )
        try:
            id = int(id)
        except (TypeError, ValueError):
            pass
        obj = model() if id is None else model.get_by_id(id)
        if id is not None and obj is None:
            raise ValidationError(
                "Object with id {id} not found".format(id=id)
            )
        return obj

    def serialize(self, serializable):
        """
        Return a serialized representation of `serializable`
        `serializable` is either a queryset or a model object
        """
        if isinstance(serializable, ndb.Model):
            return self.to_dict_repr(serializable)
        elif isinstance(serializable, ndb.Query):
            serializable = serializable.fetch()
            return [
                self.to_dict_repr(obj) for obj in serializable
            ]
        raise NonSerializableException(
            "Object of type {type} is not serializable".format(
                type=type(serializable).__name__
            )
        )

    def to_dict_repr(self, obj):
        """
        Model `Obj` -> dict representation
        Override this method to define custom dict representations
        """
        dct = obj.to_dict(
            include=self.include_fields,
            exclude=self.exclude_fields
        )
        dct['id'] = obj.key.id()
        for key, value in self.translate_fields.iteritems():
            dct[value] = dct.pop(key)
        for key in dct:
            dct[key] = getattr(self, 'get_' + key, lambda _: dct[key])(obj)
        return dct

    def update(self, data, id, model=None, partial=False):
        """
        Update the model object with at `id` with the given `data`
        Optionally take `model` as an argument (default is self.model)
        """
        return self._save(data, instance=self.get_obj(id=id, model=model))



"""
Viewsets club together list and detail views
providing a clean way of handling requests
"""
import json
import webapp2
import http
from zennla.renderers import JSONRenderer
from zennla import exceptions as zennla_exceptions


class ModelViewSet(webapp2.RequestHandler):
    """
    ModelViewSet can be used to handle requests on a resource level
    Can be subclassed to create specialized viewsets for a resource
    The attributes that can be set are:
        - `query`: The base query on which all list operations occur
        - `serializer_class`: The serializers.ModelSerializer class
                which is to be used for serialization
    """
    model = None
    serializer_class = None
    filter_backends = []
    renderers = [JSONRenderer]

    def __init__(self, *args, **kwargs):
        super(ModelViewSet, self).__init__(*args, **kwargs)
        self.media_renderer_map = {
            renderer.media_type.lower(): renderer
            for renderer in self.renderers
        }
        self.media_renderer_map.update({'*/*': self.renderers[0]})

    def dispatch(self):
        """Dispatches the request.

        This will first check if there's a handler_method defined in the
        matched route, and if not it'll use the method correspondent to the
        request method (``get()``, ``post()`` etc).
        """
        request = self.request
        method_name = request.route.handler_method or \
            request.method.lower().replace('-', '_')

        method = getattr(self, method_name, None)
        if method is None:
            # 405 Method Not Allowed.
            # The response MUST include an Allow header containing a
            # list of valid methods for the requested resource.
            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.4.6
            valid = ', '.join(self._get_handler_methods())
            self.abort(405, headers=[('Allow', valid)])

        # The handler only receives *args if no named variables are set.
        args, kwargs = request.route_args, request.route_kwargs
        if kwargs:
            args = ()
        try:
            renderer = self.get_renderer()
        except zennla_exceptions.UnacceptableRequest as e:
            renderer = self.renderers[0]()
            self.response.headers['Content-Type'] = renderer.media_type
            self.response.write(renderer.render(e.detail))
            self.response.status_int = e.status_code
            return
        try:
            pre_method_handler = getattr(self, 'pre_' + method_name, None)
            if pre_method_handler is not None:
                pre_method_handler(*args, **kwargs)
            try:
                response = method(*args, **kwargs)
            except zennla_exceptions.APIException as e:
                self.response.headers['Content-Type'] = renderer.media_type
                self.response.write(renderer.render(e.detail))
                self.response.status_int = e.status_code
                return
            post_method_handler = getattr(self, 'post_' + method_name, None)
            if post_method_handler is not None:
                post_method_handler(*args, **kwargs)
            self.response.headers['Content-Type'] = renderer.media_type
            return response
        except Exception, e:
            return self.handle_exception(e, self.app.debug)

    def filter_query(self, query):
        """
        Filter the query against any filter backends supplied
        Filter backends are listed by the `filter_backends` attribute.
        """
        filtered_query = query
        for filter_backend in self.filter_backends:
            filtered_query = filter_backend().get_filtered_query(
                filtered_query, self.request.GET
            )
        return filtered_query

    def get_model(self, *args, **kwargs):
        """
        Return the `model` for the viewset
        Override for any model switching logic
        """
        model = self.model or self.serializer_class.model
        if model is None:
            raise zennla_exceptions.ImproperlyConfigured(
                "No model associated with viewset or the serializer. "
                "Either set the `model` attribute of the serializer "
                "{serializer} or the viewset {viewset} or override "
                "`get_model()` to return an ndb model instance.".format(
                    serializer=self.serializer_class.__name__,
                    viewset=self.__class__.__name__
                )
            )
        return model

    def get_query(self, *args, **kwargs):
        """
        Return the `query` for the viewset
        Override and add any pre-filtering required on the queryset
        """
        return self.get_model().query()

    def get_serializer_class(self, *args, **kwargs):
        """
        Return the `serializer_class` for the viewset
        Override this method to dynamically select a serializer class
        based on the request
        """
        if self.serializer_class is None:
            return zennla_exceptions.ImproperlyConfigured(
                "No serializer class associated with the viewset {viewset}. "
                "Set the `serializer_class` attribute of the viewset or "
                "override `get_serializer_class()` to return a "
                "`ModelSerializer` class or subclass".format(
                    viewset=self.__class__.__name__
                )
            )
        return self.serializer_class

    def get(self, *args, **kwargs):
        """
        Correspond to HTTP GET
        """
        if args or kwargs:
            self.retrieve(*args, **kwargs)
        else:
            self.list(*args, **kwargs)

    def list(self, *args, **kwargs):
        """
        Handle GET resource-list
        """
        query = self.filter_query(self.get_query(*args, **kwargs))
        serializer = self.get_serializer_class(*args, **kwargs)()
        data = serializer.serialize(query)
        self.response.write(self.get_renderer().render(data))

    def retrieve(self, *args, **kwargs):
        """
        Handle GET resource-detail
        """
        serializer = self.get_serializer_class(*args, **kwargs)()
        obj = serializer.get_obj(id=kwargs.values()[0])
        data = serializer.serialize(obj)
        self.response.write(self.get_renderer().render(data))

    def post(self, *args, **kwargs):
        """
        Correspond to HTTP POST
        """
        data = json.loads(self.request.body)
        serializer = self.get_serializer_class(*args, **kwargs)()
        obj = serializer.create(data=data)
        self.response.write(
            self.get_renderer().render(serializer.serialize(obj))
        )

    def put(self, *args, **kwargs):
        """
        Correspond to HTTP PUT
        """
        data = json.loads(self.request.body)
        serializer = self.get_serializer_class(*args, **kwargs)()
        updated_obj = serializer.update(data=data, id=kwargs.values()[0])
        self.response.write(
            self.get_renderer().render(serializer.serialize(updated_obj))
        )

    def delete(self, *args, **kwargs):
        """
        Correspond to HTTP DELETE
        """
        serializer = self.get_serializer_class(*args, **kwargs)()
        obj = serializer.get_obj(id=kwargs.values()[0])
        obj.key.delete()
        self.response.status_int = http.HTTP_204_NO_CONTENT

    def patch(self, *args, **kwargs):
        """
        Correspond to HTTP PATCH
        """
        kwargs['partial'] = True
        self.put(*args, **kwargs)

    def get_renderer(self, *args, **kwargs):
        """
        Return the renderer to be used to render the response
        """
        request_media_type = self.request.headers.get(
            'Accept', '*/*'
        ).lower()
        renderer_class = self.media_renderer_map.get(request_media_type)
        if not renderer_class:
            raise zennla_exceptions.UnacceptableRequest(
                "Could not satisfy the request Accept header"
            )
        return renderer_class()



"""
FilterSets can be used as filtering backends for a viewset.
They perform querying against query parameters supplied in a request
"""

import operator
from google.appengine.ext.db import BadValueError
from zennla.exceptions import ImproperlyConfigured, ValidationError


class FieldFilter(object):
    """
    The base class for all field type filters
    Extend to create custom Field Filters
    """
    def __init__(self, field, lookup_type='eq'):
        """
        FieldFilter has two base attributes:
            `field`: The ndb field for which the filter is applicable
            `lookup_type`: The type of comparison against the value supplied.
                It can have the following values:
                - `in`: field value in [list, of, values]
                - `eq`: field value == value
                - `ne`: field value != value
                - `gt`: field value > value
                - `ge`: field value >= value
                - `lt`: field value < value
                - `le`: field value <= value
        """
        self.field = field
        self.lookup_type = lookup_type

    def get_converted_value(self, value):
        """
        Convert the input value to another value
        Override this method to create filters that take non-string values
        """
        return value

    def get_filter(self, value):
        """
        Return a FilterNode object comparing the `field` with the `value`
        using the `lookup_type`. You are unlikely to want to override
        this method.
        """
        try:
            if self.lookup_type == 'in':
                if not isinstance(value, (list, tuple)):
                    raise ImproperlyConfigured(
                        "IN comparison must be against a list or a tuple. "
                        "Found {type} instead".format(
                            type=type(value).__name__
                        )
                    )
                converted_value = [
                    self.get_converted_value(elem) for elem in value
                ]
                return self.field.IN(converted_value)
            try:
                converted_value = self.get_converted_value(value)
                return getattr(operator, self.lookup_type)(
                    self.field, converted_value
                )
            except AttributeError:
                raise ImproperlyConfigured(
                    "`lookup_type` must be one of: `in`, `eq`, `ne`, `gt`, "
                    "`ge`, `lt`, `le`. Found {lookup_field} instead.".format(
                        lookup_field=self.lookup_type
                    )
                )
        except BadValueError:
            raise ImproperlyConfigured(
                "Mismatch in the filter type and the model's field type "
                "({filter_type} used for {field_type})".format(
                    filter_type=type(self).__name__,
                    field_type=type(self.field).__name__
                )
            )


class NumberFilter(FieldFilter):
    """
    Filter for numeric values
    """
    def get_converted_value(self, value):
        """
        Overridden to convert input value to its numeric equivalent
        """
        try:
            return float(value)
        except ValueError:
            raise ValidationError(
                "Value passed for {filter_type} is not a number. "
                "Found {value_type} instead".format(
                    filter_type=type(self).__name__,
                    value_type=type(value).__name__
                )
            )


class BooleanFilter(FieldFilter):
    """
    Filter for boolean values
    """
    def get_converted_value(self, value):
        """
        Overridden to convert input value to its boolean equivalent
        """
        return value.lower() not in ('false', '0', 'no', 'f', 'n')


class StringFilter(FieldFilter):
    """
    Filter for string values
    """
    def get_converted_value(self, value):
        """
        Overridden to convert input value to its string equivalent
        """
        try:
            return unicode(value)
        except ValueError:
            raise ValidationError(
                "Value passed for {filter_type} is not a string. "
                "Found {value_type} instead".format(
                    filter_type=type(self).__name__,
                    value_type=type(value).__name__
                )
            )


class FilterSet(object):
    """
    A FilterSet is associated with a ViewSet. The FilterSet filters the query
    based on any FilterSet associated with it.
    """
    class Meta:
        filters = []  # List any FieldFilters that need to be applied here

    def get_filters(self, field_values):
        """
        Return a generator for all the FilterFields listed in `Meta.filters`
        `field_values` is a mapping of field names and their values
        and is used to compare against the FilterFields
        """
        return (
            getattr(self, filter_name).get_filter(
                field_values.getall(filter_name)
                if getattr(
                    self, filter_name
                ).lookup_type == 'in' else field_values.get(filter_name)
            )
            for filter_name in self.Meta.filters
            if filter_name in field_values
        )

    def get_filtered_query(self, query, field_values):
        """
        Filter the `query` against the filters listed in `Meta.filters`
        """
        return query.filter(*self.get_filters(field_values))



from distutils.core import setup
setup(
  name = 'zennla',
  packages = ['zennla'],
  version = '0.4',
  description = 'A REST microframework for Google App Engine',
  author = 'Abhinav Desor',
  author_email = 'abhinav.desor@gmail.com',
  url = 'https://github.com/adesor/zenn-la',
  download_url = 'https://github.com/adesor/zenn-la/tarball/0.4',
  keywords = ['google app engine', 'app engine', 'gae', 'ndb', 'webapp2', 'rest', 'framework'],
  classifiers = [],
)
